# -*- coding: utf-8 -*-
# Code for Life
#
# Copyright (C) 2019, Ocado Innovation Limited
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ADDITIONAL TERMS – Section 7 GNU General Public Licence
#
# This licence does not grant any right, title or interest in any “Ocado” logos,
# trade names or the trademark “Ocado” or any other trademarks or domain names
# owned by Ocado Innovation Limited or the Ocado group of companies or any other
# distinctive brand features of “Ocado” as may be secured from time to time. You
# must not distribute any modification of this program using the trademark
# “Ocado” or claim any affiliation or association with Ocado or its employees.
#
# You are not authorised to use the name Ocado (or any of its trade names) or
# the names of any author or contributor in advertising or for publicity purposes
# pertaining to the distribution of this program, without the prior written
# authorisation of Ocado.
#
# Any propagation, distribution or conveyance of this program must include this
# copyright notice and these terms. You must not misrepresent the origins of this
# program; modified versions of the program must be marked as such and not
# identified as the original program.

"""
    Theme data
"""

from rest_framework.reverse import reverse
from django.utils.translation import ugettext


class Theme(object):
    def __init__(self, pk, name, text, background, border, selected):
        self.id = self.pk = pk
        self.name = name
        self.text = text
        self.background = background
        self.border = border
        self.selected = selected


THEME_DATA = {
    "grass": Theme(
        name=u"grass",
        text=ugettext(u"Grass"),
        selected=u"#bce369",
        background=u"#eef7ff",
        border=u"#70961f",
        pk=1,
    ),
    "snow": Theme(
        name=u"snow",
        text=ugettext(u"Snow"),
        selected=u"#b3deff",
        background=u"#eef7ff",
        border=u"#83c9fe",
        pk=2,
    ),
    "farm": Theme(
        name=u"farm",
        text=ugettext(u"Farm"),
        selected=u"#bce369",
        background=u"#eef7ff",
        border=u"#70961f",
        pk=3,
    ),
    "city": Theme(
        name=u"city",
        text=ugettext(u"City"),
        selected=u"#C1C1C1",
        background=u"#eef7ff",
        border=u"#686868",
        pk=4,
    ),
}


def get_theme(name):
    """ Helper method to get a theme."""
    return THEME_DATA[name]


def get_all_themes():
    return THEME_DATA.values()


def get_theme_by_pk(pk):
    for theme in THEME_DATA.values():
        if theme.pk == int(pk):
            return theme
    raise KeyError


def get_themes_url(pk, request):
    return reverse("theme-detail", args={pk}, request=request)



# -*- coding: utf-8 -*-
# Code for Life
#
# Copyright (C) 2019, Ocado Innovation Limited
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ADDITIONAL TERMS – Section 7 GNU General Public Licence
#
# This licence does not grant any right, title or interest in any “Ocado” logos,
# trade names or the trademark “Ocado” or any other trademarks or domain names
# owned by Ocado Innovation Limited or the Ocado group of companies or any other
# distinctive brand features of “Ocado” as may be secured from time to time. You
# must not distribute any modification of this program using the trademark
# “Ocado” or claim any affiliation or association with Ocado or its employees.
#
# You are not authorised to use the name Ocado (or any of its trade names) or
# the names of any author or contributor in advertising or for publicity purposes
# pertaining to the distribution of this program, without the prior written
# authorisation of Ocado.
#
# Any propagation, distribution or conveyance of this program must include this
# copyright notice and these terms. You must not misrepresent the origins of this
# program; modified versions of the program must be marked as such and not
# identified as the original program.
from django.db import migrations


def update(apps, schema_editor):
    mappings = [
        (u"1", 1),
        (u"2", 1),
        (u"3", 1),
        (u"4", 1),
        (u"5", 1),
        (u"6", 1),
        (u"7", 1),
        (u"8", 1),
        (u"9", 1),
        (u"10", 1),
        (u"11", 1),
        (u"12", 1),
        (u"13", 2),
        (u"14", 2),
        (u"15", 2),
        (u"16", 2),
        (u"17", 2),
        (u"18", 2),
        (u"19", 3),
        (u"20", 3),
        (u"21", 3),
        (u"22", 3),
        (u"23", 3),
        (u"24", 3),
        (u"25", 3),
        (u"26", 3),
        (u"27", 3),
        (u"28", 3),
        (u"29", 4),
        (u"30", 4),
        (u"31", 4),
        (u"32", 4),
        (u"33", 5),
        (u"34", 5),
        (u"35", 5),
        (u"36", 5),
        (u"37", 5),
        (u"38", 5),
        (u"39", 5),
        (u"40", 5),
        (u"41", 5),
        (u"42", 5),
        (u"43", 5),
        (u"44", 6),
        (u"45", 6),
        (u"46", 6),
        (u"47", 6),
        (u"48", 6),
        (u"49", 6),
        (u"50", 6),
        (u"62", 7),
        (u"51", 8),
        (u"59", 8),
        (u"60", 8),
        (u"68", 9),
        (u"69", 9),
        (u"70", 9),
        (u"71", 9),
        (u"63", 7),
        (u"67", 7),
        (u"65", 7),
        (u"81", 10),
        (u"82", 10),
        (u"83", 10),
        (u"85", 10),
        (u"86", 10),
        (u"87", 10),
        (u"89", 10),
        (u"93", 11),
        (u"94", 11),
        (u"96", 11),
        (u"95", 11),
        (u"98", 11),
        (u"99", 11),
        (u"52", 8),
        (u"53", 8),
        (u"54", 8),
        (u"55", 8),
        (u"56", 8),
        (u"57", 8),
        (u"58", 8),
        (u"61", 7),
        (u"64", 7),
        (u"66", 7),
        (u"72", 9),
        (u"75", 9),
        (u"74", 9),
        (u"73", 9),
        (u"79", 9),
        (u"80", 10),
        (u"84", 10),
        (u"88", 10),
        (u"90", 10),
        (u"91", 10),
        (u"92", 11),
        (u"97", 11),
        (u"100", 11),
        (u"101", 11),
        (u"102", 11),
        (u"103", 11),
        (u"104", 11),
        (u"105", 11),
        (u"106", 11),
        (u"107", 11),
        (u"108", 11),
        (u"109", 11),
        (u"76", 9),
        (u"77", 9),
        (u"78", 9),
    ]
    for mapping in mappings:
        level = apps.get_model("game", "Level").objects.get(
            name=mapping[0], default=True
        )
        level.episode = apps.get_model("game", "Episode").objects.get(pk=mapping[1])
        level.save()


class Migration(migrations.Migration):
    dependencies = [("game", "0040_auto_20150128_2019")]

    operations = [migrations.RunPython(update)]



# -*- coding: utf-8 -*-
# Code for Life
#
# Copyright (C) 2019, Ocado Innovation Limited
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ADDITIONAL TERMS – Section 7 GNU General Public Licence
#
# This licence does not grant any right, title or interest in any “Ocado” logos,
# trade names or the trademark “Ocado” or any other trademarks or domain names
# owned by Ocado Innovation Limited or the Ocado group of companies or any other
# distinctive brand features of “Ocado” as may be secured from time to time. You
# must not distribute any modification of this program using the trademark
# “Ocado” or claim any affiliation or association with Ocado or its employees.
#
# You are not authorised to use the name Ocado (or any of its trade names) or
# the names of any author or contributor in advertising or for publicity purposes
# pertaining to the distribution of this program, without the prior written
# authorisation of Ocado.
#
# Any propagation, distribution or conveyance of this program must include this
# copyright notice and these terms. You must not misrepresent the origins of this
# program; modified versions of the program must be marked as such and not
# identified as the original program.
from django.db import migrations
from game.level_management import set_decor_inner, set_blocks_inner
import json


def new_level(apps, schema_editor):
    Level = apps.get_model("game", "Level")
    Character = apps.get_model("game", "Character")
    Theme = apps.get_model("game", "Theme")
    LevelDecor = apps.get_model("game", "LevelDecor")
    LevelBlock = apps.get_model("game", "LevelBlock")
    Block = apps.get_model("game", "Block")

    def set_decor(level, decor):
        set_decor_inner(level, decor, LevelDecor)

    def set_blocks(level, blocks):
        set_blocks_inner(level, blocks, LevelBlock, Block)

    level79 = Level.objects.get(name="79", default=1)
    level78 = Level(
        name="78",
        default=True,
        path='[{"coordinate":[2,7],"connectedNodes":[1]},{"coordinate":[3,7],"connectedNodes":[0,2]},{"coordinate":[4,7],"connectedNodes":[1,3]},{"coordinate":[5,7],"connectedNodes":[2,4]},{"coordinate":[6,7],"connectedNodes":[3,5]},{"coordinate":[7,7],"connectedNodes":[4,6]},{"coordinate":[8,7],"connectedNodes":[5,7]},{"coordinate":[9,7],"connectedNodes":[6,8]},{"coordinate":[9,6],"connectedNodes":[7,9]},{"coordinate":[9,5],"connectedNodes":[8,10]},{"coordinate":[9,4],"connectedNodes":[9,11]},{"coordinate":[9,3],"connectedNodes":[10,12]},{"coordinate":[9,2],"connectedNodes":[11,13]},{"coordinate":[9,1],"connectedNodes":[12,14]},{"coordinate":[9,0],"connectedNodes":[15,13]},{"coordinate":[8,0],"connectedNodes":[16,14]},{"coordinate":[8,1],"connectedNodes":[17,15]},{"coordinate":[8,2],"connectedNodes":[18,16]},{"coordinate":[8,3],"connectedNodes":[19,17]},{"coordinate":[8,4],"connectedNodes":[20,18]},{"coordinate":[8,5],"connectedNodes":[22,21,19]},{"coordinate":[8,6],"connectedNodes":[20]},{"coordinate":[7,5],"connectedNodes":[23,20]},{"coordinate":[6,5],"connectedNodes":[24,22]},{"coordinate":[5,5],"connectedNodes":[25,23]},{"coordinate":[4,5],"connectedNodes":[26,24]},{"coordinate":[3,5],"connectedNodes":[27,25]},{"coordinate":[2,5],"connectedNodes":[62,53,26]},{"coordinate":[1,4],"connectedNodes":[29]},{"coordinate":[1,3],"connectedNodes":[54,28,30]},{"coordinate":[1,2],"connectedNodes":[29,60]},{"coordinate":[1,0],"connectedNodes":[65,60,32]},{"coordinate":[2,0],"connectedNodes":[31,33]},{"coordinate":[3,0],"connectedNodes":[32,34]},{"coordinate":[4,0],"connectedNodes":[33,35]},{"coordinate":[5,0],"connectedNodes":[34,36]},{"coordinate":[6,0],"connectedNodes":[35,38,37]},{"coordinate":[7,0],"connectedNodes":[36]},{"coordinate":[6,1],"connectedNodes":[39,36]},{"coordinate":[6,2],"connectedNodes":[40,38]},{"coordinate":[6,3],"connectedNodes":[42,41,39]},{"coordinate":[6,4],"connectedNodes":[40]},{"coordinate":[5,3],"connectedNodes":[43,44,40,45]},{"coordinate":[4,3],"connectedNodes":[42]},{"coordinate":[5,4],"connectedNodes":[42]},{"coordinate":[5,2],"connectedNodes":[42,46]},{"coordinate":[5,1],"connectedNodes":[47,45]},{"coordinate":[4,1],"connectedNodes":[48,46]},{"coordinate":[3,1],"connectedNodes":[49,47]},{"coordinate":[2,1],"connectedNodes":[50,48]},{"coordinate":[2,2],"connectedNodes":[51,49]},{"coordinate":[2,3],"connectedNodes":[52,50]},{"coordinate":[2,4],"connectedNodes":[51]},{"coordinate":[2,6],"connectedNodes":[27]},{"coordinate":[0,3],"connectedNodes":[55,29,64]},{"coordinate":[0,4],"connectedNodes":[56,54]},{"coordinate":[0,5],"connectedNodes":[57,55]},{"coordinate":[0,6],"connectedNodes":[58,59,56]},{"coordinate":[0,7],"connectedNodes":[57]},{"coordinate":[1,6],"connectedNodes":[57,63,62]},{"coordinate":[1,1],"connectedNodes":[61,30,31]},{"coordinate":[0,1],"connectedNodes":[60]},{"coordinate":[1,5],"connectedNodes":[59,27]},{"coordinate":[1,7],"connectedNodes":[59]},{"coordinate":[0,2],"connectedNodes":[54]},{"coordinate":[0,0],"connectedNodes":[31]}]',
        traffic_lights="[]",
        destinations="[[2,3]]",
        origin='{"coordinate":[2,7],"direction":"E"}',
        max_fuel=99,
        blocklyEnabled=True,
        pythonEnabled=False,
        theme=Theme.objects.get(id=4),
        character=Character.objects.get(id="6"),
        model_solution="[12]",
        next_level=level79,
        disable_route_score=True,
    )
    level78.save()
    set_decor(
        level78,
        json.loads(
            '[{"x":453,"y":592,"decorName":"tree1"},{"x":355,"y":593,"decorName":"tree1"},{"x":729,"y":67,"decorName":"tree2"}]'
        ),
    )
    set_blocks(
        level78,
        json.loads(
            '[{"type":"move_forwards","number":2},{"type":"turn_left"},{"type":"turn_right","number":1},{"type":"turn_around","number":3},{"type":"controls_repeat_while","number":4},{"type":"controls_repeat_until","number":1},{"type":"logic_negate","number":1},{"type":"at_destination","number":1},{"type":"road_exists","number":2},{"type":"dead_end","number":2}]'
        ),
    )

    level77 = Level.objects.get(name="77", default=1)
    level77.next_level = level78
    level77.save()


class Migration(migrations.Migration):
    dependencies = [("game", "0033_recursion_level")]

    operations = [migrations.RunPython(new_level)]



# -*- coding: utf-8 -*-
# Code for Life
#
# Copyright (C) 2019, Ocado Innovation Limited
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ADDITIONAL TERMS – Section 7 GNU General Public Licence
#
# This licence does not grant any right, title or interest in any “Ocado” logos,
# trade names or the trademark “Ocado” or any other trademarks or domain names
# owned by Ocado Innovation Limited or the Ocado group of companies or any other
# distinctive brand features of “Ocado” as may be secured from time to time. You
# must not distribute any modification of this program using the trademark
# “Ocado” or claim any affiliation or association with Ocado or its employees.
#
# You are not authorised to use the name Ocado (or any of its trade names) or
# the names of any author or contributor in advertising or for publicity purposes
# pertaining to the distribution of this program, without the prior written
# authorisation of Ocado.
#
# Any propagation, distribution or conveyance of this program must include this
# copyright notice and these terms. You must not misrepresent the origins of this
# program; modified versions of the program must be marked as such and not
# identified as the original program.
from unittest import expectedFailure, skip

from .base_game_test import BaseGameTest


def complete_and_check_level(
    level_number,
    page,
    next_episode=None,
    check_algorithm_score=True,
    check_route_score=True,
    final_level=False,
):
    page.solution_button().run_program().assert_success()
    if check_algorithm_score:
        page.assert_algorithm_score(10)
    if check_route_score:
        page.assert_route_score(10)
    if final_level:
        return page
    if next_episode is None:
        page.next_level()
        page.assert_level_number(level_number + 1)
    else:
        page.next_episode()
        page.assert_episode_number(next_episode)
    return page


class TestPlayThrough(BaseGameTest):
    def setUp(self):
        self.login_once()

    def _complete_episode(self, episode_number, level_number, **kwargs):
        page = self.go_to_episode(episode_number)
        complete_and_check_level(level_number, page, **kwargs)

    def _complete_level(self, level_number, **kwargs):
        page = self.go_to_level(level_number)
        complete_and_check_level(level_number, page, **kwargs)

    def test_episode_01(self):
        self._complete_episode(1, 1)

    def test_level_001(self):
        self._complete_level(1)

    def test_level_002(self):
        self._complete_level(2)

    def test_level_003(self):
        self._complete_level(3)

    def test_level_004(self):
        self._complete_level(4)

    def test_level_005(self):
        self._complete_level(5)

    def test_level_006(self):
        self._complete_level(6)

    def test_level_007(self):
        self._complete_level(7)

    def test_level_008(self):
        self._complete_level(8)

    def test_level_009(self):
        self._complete_level(9)

    def test_level_010(self):
        self._complete_level(10)

    def test_level_011(self):
        self._complete_level(11)

    def test_level_012(self):
        self._complete_level(12, next_episode=2)

    def test_episode_02(self):
        self._complete_episode(2, 13)

    def test_level_013(self):
        self._complete_level(13)

    def test_level_014(self):
        self._complete_level(14)

    def test_level_015(self):
        self._complete_level(15)

    def test_level_016(self):
        self._complete_level(16)

    def test_level_017(self):
        self._complete_level(17)

    def test_level_018(self):
        self._complete_level(18, next_episode=3)

    def test_episode_03(self):
        self._complete_episode(3, 19)

    def test_level_019(self):
        self._complete_level(19)

    def test_level_020(self):
        self._complete_level(20)

    def test_level_021(self):
        self._complete_level(21)

    def test_level_022(self):
        self._complete_level(22)

    def test_level_023(self):
        self._complete_level(23)

    def test_level_024(self):
        self._complete_level(24)

    def test_level_025(self):
        self._complete_level(25)

    def test_level_026(self):
        self._complete_level(26)

    def test_level_027(self):
        self._complete_level(27)

    def test_level_028(self):
        self._complete_level(28, next_episode=4)

    def test_episode_04(self):
        self._complete_episode(4, 29)

    def test_level_029(self):
        self._complete_level(29)

    def test_level_030(self):
        self._complete_level(30)

    def test_level_031(self):
        self._complete_level(31)

    def test_level_032(self):
        self._complete_level(32, next_episode=5)

    def test_episode_05(self):
        self._complete_episode(5, 33)

    def test_level_033(self):
        self._complete_level(33)

    def test_level_034(self):
        self._complete_level(34)

    def test_level_035(self):
        self._complete_level(35)

    def test_level_036(self):
        self._complete_level(36)

    def test_level_037(self):
        self._complete_level(37)

    def test_level_038(self):
        self._complete_level(38)

    def test_level_039(self):
        self._complete_level(39)

    def test_level_040(self):
        self._complete_level(40)

    def test_level_041(self):
        self._complete_level(41)

    def test_level_042(self):
        self._complete_level(42)

    def test_level_043(self):
        self._complete_level(43, next_episode=6)

    def test_episode_06(self):
        self._complete_episode(6, 44)

    def test_level_044(self):
        self._complete_level(44)

    def test_level_045(self):
        self._complete_level(45)

    def test_level_046(self):
        self._complete_level(46)

    def test_level_047(self):
        self._complete_level(47)

    def test_level_048(self):
        self._complete_level(48)

    def test_level_049(self):
        self._complete_level(49)

    def test_level_050(self):
        self._complete_level(50, next_episode=7)

    def test_episode_07(self):
        self._complete_episode(7, 51, check_algorithm_score=False)

    def test_level_051(self):
        self._complete_level(51, check_algorithm_score=False)

    def test_level_052(self):
        self._complete_level(52, check_route_score=False)

    def test_level_053(self):
        self._complete_level(53)

    def test_level_054(self):
        self._complete_level(54, check_route_score=False)

    def test_level_055(self):
        self._complete_level(55)

    def test_level_056(self):
        self._complete_level(56)

    def test_level_057(self):
        self._complete_level(57)

    def test_level_058(self):
        self._complete_level(58)

    def test_level_059(self):
        self._complete_level(59, check_route_score=False)

    def test_level_060(self):
        self._complete_level(60, next_episode=8, check_route_score=False)

    def test_episode_08(self):
        self._complete_episode(8, 61)

    def test_level_061(self):
        self._complete_level(61)

    def test_level_062(self):
        self._complete_level(62)

    def test_level_063(self):
        self._complete_level(63)

    def test_level_064(self):
        self._complete_level(64)

    def test_level_065(self):
        self._complete_level(65)

    def test_level_066(self):
        self._complete_level(66)

    def test_level_067(self):
        self._complete_level(67, next_episode=9)

    def test_episode_09(self):
        self._complete_episode(9, 68, check_route_score=False)

    def test_level_068(self):
        self._complete_level(68, check_route_score=False)

    def test_level_069(self):
        self._complete_level(69, check_route_score=False)

    def test_level_070(self):
        self._complete_level(70, check_route_score=False)

    def test_level_071(self):
        self._complete_level(71)

    def test_level_072(self):
        self._complete_level(72)

    def test_level_073(self):
        self._complete_level(73)

    def test_level_074(self):
        self._complete_level(74, check_route_score=False)

    def test_level_075(self):
        self._complete_level(75)

    def test_level_076(self):
        self._complete_level(76, check_route_score=False)

    def test_level_077(self):
        self._complete_level(77, check_route_score=False)

    def test_level_078(self):
        self._complete_level(78, check_route_score=False)

    def test_level_079(self):
        self._complete_level(79, next_episode=10)

    def test_episode_10(self):
        self._complete_episode(10, 80, check_algorithm_score=False)

    def test_level_080(self):
        self._complete_level(80, check_algorithm_score=False)

    def test_level_081(self):
        self._complete_level(81, check_algorithm_score=False)

    def test_level_082(self):
        self._complete_level(82, check_algorithm_score=False)

    def test_level_083(self):
        self._complete_level(83, check_algorithm_score=False)

    def test_level_084(self):
        self._complete_level(84, check_algorithm_score=False)

    def test_level_085(self):
        self._complete_level(85, check_algorithm_score=False)

    def test_level_086(self):
        self._complete_level(86, check_algorithm_score=False)

    def test_level_087(self):
        self._complete_level(87, check_algorithm_score=False)

    def test_level_088(self):
        self._complete_level(88, check_algorithm_score=False)

    def test_level_089(self):
        self._complete_level(89, check_algorithm_score=False)

    def test_level_090(self):
        self._complete_level(90, check_algorithm_score=False)

    def test_level_091(self):
        self._complete_level(91, next_episode=11, check_algorithm_score=False)

    def test_episode_11(self):
        self._complete_episode(11, 92, check_algorithm_score=False)

    def test_level_092(self):
        self._complete_level(92, check_algorithm_score=False)

    def test_level_093(self):
        self._complete_level(93, check_algorithm_score=False)

    def test_level_094(self):
        self._complete_level(94, check_algorithm_score=False)

    def test_level_095(self):
        self._complete_level(95, check_algorithm_score=False)

    def test_level_096(self):
        self._complete_level(96, check_algorithm_score=False)

    def test_level_097(self):
        self._complete_level(97, check_algorithm_score=False)

    def test_level_098(self):
        self._complete_level(98, check_algorithm_score=False)

    def test_level_099(self):
        self._complete_level(99, check_algorithm_score=False)

    def test_level_100(self):
        self._complete_level(100, check_algorithm_score=False)

    def test_level_101(self):
        self._complete_level(101, check_algorithm_score=False)

    def test_level_102(self):
        self._complete_level(102, check_algorithm_score=False)

    def test_level_103(self):
        self._complete_level(103, check_algorithm_score=False)

    def test_level_104(self):
        self._complete_level(104, check_algorithm_score=False)

    def test_level_105(self):
        self._complete_level(105, check_algorithm_score=False)

    def test_level_106(self):
        self._complete_level(106, check_algorithm_score=False)

    def test_level_107(self):
        self._complete_level(107, check_algorithm_score=False)

    def test_level_108(self):
        self._complete_level(108, check_algorithm_score=False)

    def test_level_109(self):
        self._complete_level(109, check_algorithm_score=False, final_level=True)



# -*- coding: utf-8 -*-
# Code for Life
#
# Copyright (C) 2019, Ocado Innovation Limited
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ADDITIONAL TERMS – Section 7 GNU General Public Licence
#
# This licence does not grant any right, title or interest in any “Ocado” logos,
# trade names or the trademark “Ocado” or any other trademarks or domain names
# owned by Ocado Innovation Limited or the Ocado group of companies or any other
# distinctive brand features of “Ocado” as may be secured from time to time. You
# must not distribute any modification of this program using the trademark
# “Ocado” or claim any affiliation or association with Ocado or its employees.
#
# You are not authorised to use the name Ocado (or any of its trade names) or
# the names of any author or contributor in advertising or for publicity purposes
# pertaining to the distribution of this program, without the prior written
# authorisation of Ocado.
#
# Any propagation, distribution or conveyance of this program must include this
# copyright notice and these terms. You must not misrepresent the origins of this
# program; modified versions of the program must be marked as such and not
# identified as the original program.
from game.end_to_end_tests.base_game_test import BaseGameTest
from .test_play_through import complete_and_check_level


class TestCrashes(BaseGameTest):
    def test_crash_turning_left_on_straight_road(self):
        self.run_crashing_test(
            level=6, workspace_file="crash_turning_left_on_straight_road"
        )

    def test_crash_turning_left_on_right_turn(self):
        self.run_crashing_test(
            level=6, workspace_file="crash_turning_left_on_right_turn"
        )

    def test_crash_turning_right_on_straight_road(self):
        self.run_crashing_test(
            level=6, workspace_file="crash_turning_right_on_straight_road"
        )

    def test_crash_turning_right_on_left_turn(self):
        self.run_crashing_test(
            level=6, workspace_file="crash_turning_right_on_left_turn"
        )

    def test_crash_going_forward_on_right_turn(self):
        self.run_crashing_test(
            level=6, workspace_file="crash_going_forward_on_right_turn"
        )

    def test_crash_going_forward_on_left_turn(self):
        self.run_crashing_test(
            level=6, workspace_file="crash_going_forward_on_left_turn"
        )

    def test_crash_going_forward_on_t_junction(self):
        self.run_crashing_test(
            level=13, workspace_file="crash_going_forward_on_t_junction"
        )

    def test_crash_going_left_on_t_junction(self):
        self.run_crashing_test(
            level=40, workspace_file="crash_going_left_on_t_junction"
        )

    def test_crash_going_right_on_t_junction(self):
        self.run_crashing_test(
            level=40, workspace_file="crash_going_right_on_t_junction"
        )

    def test_running_out_of_instructions(self):
        self.running_out_of_instructions_test(
            level=6, workspace_file="running_out_of_instructions"
        )

    def test_running_out_of_fuel(self):
        self.running_out_of_fuel_test(level=68, workspace_file="running_out_of_fuel")

    def test_running_a_red_light(self):
        self.running_a_red_light_test(level=44, workspace_file="running_a_red_light")

    def test_not_delivered_everywhere(self):
        self.not_delivered_everywhere_test(
            level=16, workspace_file="not_delivered_everywhere"
        )

    def test_procedure_undefined(self):
        self.undefined_procedure_test(level=63, workspace_file="undefined_procedure")

    def test_crash_turning_left_clear_and_pass_level(self):
        page = (
            self.run_crashing_test(
                level=6, workspace_file="crash_turning_left_on_straight_road"
            )
            .try_again()
            .clear()
        )

        complete_and_check_level(6, page)



# -*- coding: utf-8 -*-
# Code for Life
#
# Copyright (C) 2019, Ocado Innovation Limited
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ADDITIONAL TERMS – Section 7 GNU General Public Licence
#
# This licence does not grant any right, title or interest in any “Ocado” logos,
# trade names or the trademark “Ocado” or any other trademarks or domain names
# owned by Ocado Innovation Limited or the Ocado group of companies or any other
# distinctive brand features of “Ocado” as may be secured from time to time. You
# must not distribute any modification of this program using the trademark
# “Ocado” or claim any affiliation or association with Ocado or its employees.
#
# You are not authorised to use the name Ocado (or any of its trade names) or
# the names of any author or contributor in advertising or for publicity purposes
# pertaining to the distribution of this program, without the prior written
# authorisation of Ocado.
#
# Any propagation, distribution or conveyance of this program must include this
# copyright notice and these terms. You must not misrepresent the origins of this
# program; modified versions of the program must be marked as such and not
# identified as the original program.
from django.test import TestCase

import json
import random


def find_node(x, y, nodes):
    for node in nodes:
        if node["coordinate"][0] == x and node["coordinate"][1] == y:
            return True
    return False


def check_direction(node, neighbour):
    if neighbour["coordinate"][1] < node["coordinate"][1]:
        direction = "N"
    elif neighbour["coordinate"][0] < node["coordinate"][0]:
        direction = "E"
    elif neighbour["coordinate"][1] > node["coordinate"][1]:
        direction = "S"
    elif neighbour["coordinate"][0] > node["coordinate"][0]:
        direction = "W"
    return direction


class RandomRoadTestCase(TestCase):
    def create_test_data(
        self,
        num_tiles=None,
        branchiness=None,
        loopiness=None,
        curviness=None,
        traffic_lights_enabled=False,
        decor_enabled=True,
        cows_enabled=False,
    ):
        if num_tiles is None:
            num_tiles = random.randint(3, 40)
        if branchiness is None:
            branchiness = random.randint(0, 10)
        if loopiness is None:
            loopiness = random.randint(0, 10)
        if curviness is None:
            curviness = random.randint(0, 10)

        from game.random_road import generate_random_map_data

        return generate_random_map_data(
            num_tiles,
            branchiness,
            loopiness,
            curviness,
            traffic_lights_enabled,
            decor_enabled,
            cows_enabled,
        )

    def check_if_valid(self, origin, destinations, path, decor):
        check = find_node(origin["coordinate"][0], origin["coordinate"][1], path)
        self.assertTrue(check)

        for node in destinations:
            self.assertTrue(find_node(node[0], node[1], path))

        if decor:
            # Check if any decor got generated
            self.assertTrue(decor)

    def test_tiles_less_or_equal_size(self):
        """ Test that the number of tiles is less or equal to the size parameter"""

        number_of_tiles = 30
        data = self.create_test_data(num_tiles=number_of_tiles)

        path = json.loads(data["path"])
        destinations = json.loads(data["destinations"])
        origin = json.loads(data["origin"])
        decor = data["decor"]

        self.check_if_valid(origin, destinations, path, decor)

        # Test if number of tiles = size
        self.assertTrue(len(path) <= number_of_tiles)

    def test_branchiness_min(self):
        """ Test that if the branchiness is 0 we don't get branches. """

        data = self.create_test_data(branchiness=0, loopiness=0)

        path = json.loads(data["path"])
        destinations = json.loads(data["destinations"])
        origin = json.loads(data["origin"])
        decor = data["decor"]

        self.check_if_valid(origin, destinations, path, decor)

        # Test if no branches
        for node in path:
            self.assertTrue(len(node["connectedNodes"]) <= 2)

    def test_branchiness_max(self):
        """ Test that if the branchiness > 0 we get branches. """

        data = self.create_test_data(branchiness=2, loopiness=0)

        path = json.loads(data["path"])
        destinations = json.loads(data["destinations"])
        origin = json.loads(data["origin"])
        decor = data["decor"]

        self.check_if_valid(origin, destinations, path, decor)

        # Test if some branches
        branch_count = 0
        for node in path:
            if len(node["connectedNodes"]) > 2:
                branch_count += 1

        self.assertTrue(branch_count)

    def test_curviness_max(self):
        """ Test that the path is curving significantly if the curviness is set to max value. """

        data = self.create_test_data(curviness=1)

        path = json.loads(data["path"])
        destinations = json.loads(data["destinations"])
        origin = json.loads(data["origin"])
        decor = data["decor"]

        self.check_if_valid(origin, destinations, path, decor)

        turn_count = 0
        for node in path:
            if len(node["connectedNodes"]) > 1:
                turn_count += int(
                    not (
                        check_direction(node, path[int(node["connectedNodes"][0])])
                        == check_direction(node, path[int(node["connectedNodes"][1])])
                    )
                )

        # Check if there are any turns.
        self.assertTrue(turn_count)

    def test_traffic_light_no(self):
        """ Test that traffic lights don't get generated """

        data = self.create_test_data(traffic_lights_enabled=False)

        path = json.loads(data["path"])
        destinations = json.loads(data["destinations"])
        traffic_lights = json.loads(data["traffic_lights"])
        origin = json.loads(data["origin"])
        decor = data["decor"]

        self.check_if_valid(origin, destinations, path, decor)

        # Assert no traffic lights got generated.
        self.assertFalse(traffic_lights)

    def test_traffic_light_yes(self):
        """ Test that traffic lights get generated """

        data = self.create_test_data(traffic_lights_enabled=True)

        path = json.loads(data["path"])
        destinations = json.loads(data["destinations"])
        traffic_lights = json.loads(data["traffic_lights"])
        origin = json.loads(data["origin"])
        decor = data["decor"]

        self.check_if_valid(origin, destinations, path, decor)

        # Assert no traffic lights got generated.
        self.assertTrue(traffic_lights)

    def test_decor_no(self):
        """ Test that scenery doesn't get generated """

        data = self.create_test_data(decor_enabled=False)

        path = json.loads(data["path"])
        destinations = json.loads(data["destinations"])
        origin = json.loads(data["origin"])
        decor = data["decor"]

        self.check_if_valid(origin, destinations, path, decor)

        # Assert no scenery got generated.
        self.assertFalse(decor)

    def test_decor_yes(self):
        """ Test that scenery does get generated """

        data = self.create_test_data(decor_enabled=True)

        path = json.loads(data["path"])
        destinations = json.loads(data["destinations"])
        origin = json.loads(data["origin"])
        decor = data["decor"]

        self.check_if_valid(origin, destinations, path, decor)

        # Assert scenery got generated.
        self.assertTrue(decor)



# -*- coding: utf-8 -*-
# Code for Life
#
# Copyright (C) 2019, Ocado Innovation Limited
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ADDITIONAL TERMS – Section 7 GNU General Public Licence
#
# This licence does not grant any right, title or interest in any “Ocado” logos,
# trade names or the trademark “Ocado” or any other trademarks or domain names
# owned by Ocado Innovation Limited or the Ocado group of companies or any other
# distinctive brand features of “Ocado” as may be secured from time to time. You
# must not distribute any modification of this program using the trademark
# “Ocado” or claim any affiliation or association with Ocado or its employees.
#
# You are not authorised to use the name Ocado (or any of its trade names) or
# the names of any author or contributor in advertising or for publicity purposes
# pertaining to the distribution of this program, without the prior written
# authorisation of Ocado.
#
# Any propagation, distribution or conveyance of this program must include this
# copyright notice and these terms. You must not misrepresent the origins of this
# program; modified versions of the program must be marked as such and not
# identified as the original program.
import random
import string
from django.core import mail
import sys
import email

from portal.models import Teacher, School
from portal.helpers.emails import generate_token


def generate_details(
    title="Mr", first_name="Test", last_name="Teacher", password="Password1"
):
    email_address = "testteacher%d@codeforlife.com" % random.randint(1, sys.maxint)
    return title, first_name, last_name, email_address, password


def signup_teacher_directly():
    title, first_name, last_name, email_address, password = generate_details()
    teacher = Teacher.objects.factory(
        title, first_name, last_name, email_address, password
    )
    generate_token(teacher.user.user, preverified=True)
    teacher.user.save()
    return teacher, email_address, password


def signup_teacher(page):
    page = page.go_to_teach_page()

    title, first_name, last_name, email_address, password = generate_details()
    page = page.signup(title, first_name, last_name, email_address, password, password)

    page = page.return_to_home_page()

    page = email.follow_verify_email_link_to_teach(page, mail.outbox[0])
    mail.outbox = []

    return page, email_address, password


def create_school():

    school = School()
    school.name = "".join(random.choice(string.ascii_uppercase) for _ in range(10))
    school.postcode = "".join(random.choice(string.ascii_uppercase) for _ in range(7))
    school.country = "United Kingdom"
    school.save()

    return school


def add_teacher_to_school(teacher, school):

    teacher.user.teacher.school = school
    teacher.user.teacher.save()



# -*- coding: utf-8 -*-
# Code for Life
#
# Copyright (C) 2019, Ocado Limited
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ADDITIONAL TERMS – Section 7 GNU General Public Licence
#
# This licence does not grant any right, title or interest in any “Ocado” logos,
# trade names or the trademark “Ocado” or any other trademarks or domain names
# owned by Ocado Innovation Limited or the Ocado group of companies or any other
# distinctive brand features of “Ocado” as may be secured from time to time. You
# must not distribute any modification of this program using the trademark
# “Ocado” or claim any affiliation or association with Ocado or its employees.
#
# You are not authorised to use the name Ocado (or any of its trade names) or
# the names of any author or contributor in advertising or for publicity purposes
# pertaining to the distribution of this program, without the prior written
# authorisation of Ocado.
#
# Any propagation, distribution or conveyance of this program must include this
# copyright notice and these terms. You must not misrepresent the origins of this
# program; modified versions of the program must be marked as such and not
# identified as the original program.
import os

from django.core.urlresolvers import reverse
from django.test import modify_settings

from rest_framework import status
from rest_framework.test import APITestCase

from hamcrest import *
from hamcrest.core.base_matcher import BaseMatcher

from game.decor import get_all_decor
from game.tests.utils.locale import add_new_language
from game.theme import get_all_themes
from game.character import get_all_character


class APITests(APITestCase):
    def test_list_decors(self):
        url = reverse("decor-list")
        response = self.client.get(url)
        assert_that(response, has_status_code(status.HTTP_200_OK))
        assert_that(response.data, has_length(len(get_all_decor())))

    def test_known_decor_detail(self):
        decor_id = 1
        url = reverse("decor-detail", kwargs={"pk": decor_id})
        response = self.client.get(url)
        assert_that(response, has_status_code(status.HTTP_200_OK))
        assert_that(response.data["id"], equal_to(decor_id))

    def test_unknown_decor_detail(self):
        decor_id = 0
        url = reverse("decor-detail", kwargs={"pk": decor_id})
        response = self.client.get(url)
        assert_that(response, has_status_code(status.HTTP_404_NOT_FOUND))

    def test_levels_for_known_episode(self):
        episode_id = 1
        url = reverse("level-for-episode", kwargs={"pk": episode_id})
        response = self.client.get(url)
        assert_that(response, has_status_code(status.HTTP_200_OK))
        assert_that(response.data, has_length(greater_than(0)))

    def test_levels_for_unknown_episode(self):
        episode_id = 0
        url = reverse("level-for-episode", kwargs={"pk": episode_id})
        response = self.client.get(url)
        assert_that(response, has_status_code(status.HTTP_200_OK))
        assert_that(response.data, has_length(0))

    def test_list_themes(self):
        url = reverse("theme-list")
        response = self.client.get(url)
        assert_that(response, has_status_code(status.HTTP_200_OK))
        assert_that(response.data, has_length(len(get_all_themes())))

    def test_known_theme_detail(self):
        theme_id = 1
        url = reverse("theme-detail", kwargs={"pk": theme_id})
        response = self.client.get(url)
        assert_that(response, has_status_code(status.HTTP_200_OK))
        assert_that(response.data["id"], equal_to(theme_id))

    def test_unknown_theme_detail(self):
        theme_id = 0
        url = reverse("theme-detail", kwargs={"pk": theme_id})
        response = self.client.get(url)
        assert_that(response, has_status_code(status.HTTP_404_NOT_FOUND))

    def test_list_characters(self):
        url = reverse("character-list")
        response = self.client.get(url)
        assert_that(response, has_status_code(status.HTTP_200_OK))
        assert_that(response.data, has_length(len(get_all_character())))

    def test_known_character_detail(self):
        character_id = 1
        url = reverse("character-detail", kwargs={"pk": character_id})
        response = self.client.get(url)
        assert_that(response, has_status_code(status.HTTP_200_OK))
        assert_that(response.data["id"], equal_to(character_id))

    def test_unknown_character_detail(self):
        character_id = 0
        url = reverse("character-detail", kwargs={"pk": character_id})
        response = self.client.get(url)
        assert_that(response, has_status_code(status.HTTP_404_NOT_FOUND))

    def test_list_episodes(self):
        url = reverse("episode-list")
        response = self.client.get(url)
        assert_that(response, has_status_code(status.HTTP_200_OK))
        assert_that(response.data[0]["name"], equal_to("Getting Started"))

    def test_list_episodes_with_translated_episode_names(self):
        with add_new_language():
            url = reverse("episode-list")
            response = self.client.get(url, **{"HTTP_ACCEPT_LANGUAGE": "foo-br"})
            assert_that(response, has_status_code(status.HTTP_200_OK))
            assert_that(response.data[0]["name"], equal_to("crwdns4197:0crwdne4197:0"))

    def test_episode_details(self):
        episode_id = 1
        url = reverse("episode-detail", kwargs={"pk": episode_id})
        response = self.client.get(url)
        assert_that(response, has_status_code(status.HTTP_200_OK))
        assert_that(response.data["name"], equal_to("Getting Started"))

    def test_episode_details_with_translated_episode_name(self):
        with add_new_language():
            episode_id = 1
            url = reverse("episode-detail", kwargs={"pk": episode_id})
            response = self.client.get(url, **{"HTTP_ACCEPT_LANGUAGE": "foo-br"})
            assert_that(response, has_status_code(status.HTTP_200_OK))
            assert_that(response.data["name"], equal_to("crwdns4197:0crwdne4197:0"))


def has_status_code(status_code):
    return HasStatusCode(status_code)


class HasStatusCode(BaseMatcher):
    def __init__(self, status_code):
        self.status_code = status_code

    def _matches(self, response):
        return response.status_code == self.status_code

    def describe_to(self, description):
        description.append_text("has status code ").append_text(self.status_code)

    def describe_mismatch(self, response, mismatch_description):
        mismatch_description.append_text("had status code ").append_text(
            response.status_code
        )



"""
This SeleniumTestCase is copied over from django-selenium-clean==0.3.2

Instead of inheriting from StaticLiveServerTestCase, we inherit from LiveServerTestCase.
This solves a bug introduced when upgrading to Django 1.11,
see more information here: https://github.com/jazzband/django-pipeline/issues/593
"""

from django.conf import settings
from django.contrib.staticfiles.testing import LiveServerTestCase
from django_selenium_clean import SeleniumWrapper, PageElement
from django.core.management import call_command


class SeleniumTestCase(LiveServerTestCase):

    @classmethod
    def setUpClass(cls):
        # this line is not part of django-selenium-clean
        call_command('collectstatic', '--noinput')

        super(SeleniumTestCase, cls).setUpClass()
        cls.selenium = SeleniumWrapper()
        PageElement.selenium = cls.selenium

        # Normally we would just do something like
        #     selenium.live_server_url = self.live_server_url
        # However, there is no "self" at this time, so we
        # essentially duplicate the code from the definition of
        # the LiveServerTestCase.live_server_url property.
        cls.selenium.live_server_url = 'http://%s:%s' % (
            cls.server_thread.host, cls.server_thread.port)

    @classmethod
    def tearDownClass(cls):

        cls.selenium.quit()
        PageElement.selenium = None
        super(SeleniumTestCase, cls).tearDownClass()

    def __call__(self, result=None):
        if hasattr(self, 'selenium'):
            for width in getattr(settings, 'SELENIUM_WIDTHS', [1024]):
                self.selenium.set_window_size(width, 1024)
        return super(SeleniumTestCase, self).__call__(result)



# -*- coding: utf-8 -*-
# Code for Life
#
# Copyright (C) 2019, Ocado Innovation Limited
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ADDITIONAL TERMS – Section 7 GNU General Public Licence
#
# This licence does not grant any right, title or interest in any “Ocado” logos,
# trade names or the trademark “Ocado” or any other trademarks or domain names
# owned by Ocado Innovation Limited or the Ocado group of companies or any other
# distinctive brand features of “Ocado” as may be secured from time to time. You
# must not distribute any modification of this program using the trademark
# “Ocado” or claim any affiliation or association with Ocado or its employees.
#
# You are not authorised to use the name Ocado (or any of its trade names) or
# the names of any author or contributor in advertising or for publicity purposes
# pertaining to the distribution of this program, without the prior written
# authorisation of Ocado.
#
# Any propagation, distribution or conveyance of this program must include this
# copyright notice and these terms. You must not misrepresent the origins of this
# program; modified versions of the program must be marked as such and not
# identified as the original program.

"""
    Decor data
"""

from rest_framework.reverse import reverse
from game.theme import get_theme, get_all_themes


class Decor(object):
    def __init__(self, pk, name, url, width, height, theme, z_index):
        self.id = self.pk = pk
        self.name = name
        self.url = url
        self.width = width
        self.height = height
        self.theme = theme
        self.z_index = z_index


DECOR_DATA = {
    (u"tree1", u"grass"): Decor(
        z_index=4,
        name=u"tree1",
        url=u"decor/grass/tree1.svg",
        height=100,
        width=100,
        theme=get_theme("grass"),
        pk=1,
    ),
    (u"tree2", u"grass"): Decor(
        z_index=4,
        name=u"tree2",
        url=u"decor/grass/tree2.svg",
        height=100,
        width=100,
        theme=get_theme("grass"),
        pk=2,
    ),
    (u"bush", u"grass"): Decor(
        z_index=3,
        name=u"bush",
        url=u"decor/grass/bush.svg",
        height=50,
        width=50,
        theme=get_theme("grass"),
        pk=3,
    ),
    (u"house", u"grass"): Decor(
        z_index=1,
        name=u"house",
        url=u"decor/grass/house.svg",
        height=50,
        width=50,
        theme=get_theme("grass"),
        pk=4,
    ),
    (u"cfc", u"grass"): Decor(
        z_index=1,
        name=u"cfc",
        url=u"decor/grass/cfc.svg",
        height=107,
        width=100,
        theme=get_theme("grass"),
        pk=5,
    ),
    (u"pond", u"grass"): Decor(
        z_index=2,
        name=u"pond",
        url=u"decor/grass/pond.svg",
        height=100,
        width=150,
        theme=get_theme("grass"),
        pk=6,
    ),
    (u"tree1", u"snow"): Decor(
        z_index=4,
        name=u"tree1",
        url=u"decor/snow/tree1.svg",
        height=100,
        width=100,
        theme=get_theme("snow"),
        pk=7,
    ),
    (u"tree2", u"snow"): Decor(
        z_index=4,
        name=u"tree2",
        url=u"decor/snow/tree2.svg",
        height=100,
        width=100,
        theme=get_theme("snow"),
        pk=8,
    ),
    (u"bush", u"snow"): Decor(
        z_index=3,
        name=u"bush",
        url=u"decor/snow/bush.svg",
        height=50,
        width=50,
        theme=get_theme("snow"),
        pk=9,
    ),
    (u"house", u"snow"): Decor(
        z_index=1,
        name=u"house",
        url=u"decor/snow/house.svg",
        height=50,
        width=50,
        theme=get_theme("snow"),
        pk=10,
    ),
    (u"cfc", u"snow"): Decor(
        z_index=1,
        name=u"cfc",
        url=u"decor/snow/cfc.svg",
        height=107,
        width=100,
        theme=get_theme("snow"),
        pk=11,
    ),
    (u"pond", u"snow"): Decor(
        z_index=2,
        name=u"pond",
        url=u"decor/snow/pond.svg",
        height=100,
        width=150,
        theme=get_theme("snow"),
        pk=12,
    ),
    (u"tile1", u"grass"): Decor(
        z_index=0,
        name=u"tile1",
        url=u"decor/grass/tile1.svg",
        height=100,
        width=100,
        theme=get_theme("grass"),
        pk=13,
    ),
    (u"tile1", u"snow"): Decor(
        z_index=0,
        name=u"tile1",
        url=u"decor/snow/tile1.svg",
        height=100,
        width=100,
        theme=get_theme("snow"),
        pk=14,
    ),
    (u"tile2", u"snow"): Decor(
        z_index=0,
        name=u"tile2",
        url=u"decor/snow/tile2.svg",
        height=100,
        width=100,
        theme=get_theme("snow"),
        pk=15,
    ),
    (u"house", u"farm"): Decor(
        z_index=1,
        name=u"house",
        url=u"decor/farm/house1.svg",
        height=224,
        width=184,
        theme=get_theme("farm"),
        pk=16,
    ),
    (u"cfc", u"farm"): Decor(
        z_index=1,
        name=u"cfc",
        url=u"decor/farm/cfc.svg",
        height=301,
        width=332,
        theme=get_theme("farm"),
        pk=17,
    ),
    (u"bush", u"farm"): Decor(
        z_index=3,
        name=u"bush",
        url=u"decor/farm/bush.svg",
        height=30,
        width=50,
        theme=get_theme("farm"),
        pk=18,
    ),
    (u"tree1", u"farm"): Decor(
        z_index=4,
        name=u"tree1",
        url=u"decor/farm/tree1.svg",
        height=100,
        width=100,
        theme=get_theme("farm"),
        pk=19,
    ),
    (u"tree2", u"farm"): Decor(
        z_index=4,
        name=u"tree2",
        url=u"decor/farm/house2.svg",
        height=88,
        width=65,
        theme=get_theme("farm"),
        pk=20,
    ),
    (u"pond", u"farm"): Decor(
        z_index=2,
        name=u"pond",
        url=u"decor/farm/crops.svg",
        height=100,
        width=150,
        theme=get_theme("farm"),
        pk=21,
    ),
    (u"tile1", u"farm"): Decor(
        z_index=0,
        name=u"tile1",
        url=u"decor/farm/tile1.svg",
        height=337,
        width=194,
        theme=get_theme("farm"),
        pk=22,
    ),
    (u"tile1", u"city"): Decor(
        z_index=0,
        name=u"tile1",
        url=u"decor/city/pavementTile.png",
        height=100,
        width=100,
        theme=get_theme("city"),
        pk=23,
    ),
    (u"house", u"city"): Decor(
        z_index=1,
        name=u"house",
        url=u"decor/city/house.svg",
        height=50,
        width=50,
        theme=get_theme("city"),
        pk=24,
    ),
    (u"bush", u"city"): Decor(
        z_index=3,
        name=u"bush",
        url=u"decor/city/bush.svg",
        height=50,
        width=50,
        theme=get_theme("city"),
        pk=25,
    ),
    (u"tree1", u"city"): Decor(
        z_index=4,
        name=u"tree1",
        url=u"decor/city/shop.svg",
        height=70,
        width=70,
        theme=get_theme("city"),
        pk=26,
    ),
    (u"tree2", u"city"): Decor(
        z_index=4,
        name=u"tree2",
        url=u"decor/city/school.svg",
        height=100,
        width=100,
        theme=get_theme("city"),
        pk=27,
    ),
    (u"pond", u"city"): Decor(
        z_index=2,
        name=u"pond",
        url=u"decor/city/hospital.svg",
        height=157,
        width=140,
        theme=get_theme("city"),
        pk=28,
    ),
    (u"cfc", u"city"): Decor(
        z_index=1,
        name=u"cfc",
        url=u"decor/grass/cfc.svg",
        height=107,
        width=100,
        theme=get_theme("city"),
        pk=29,
    ),
    (u"tile2", u"grass"): Decor(
        z_index=0,
        name=u"tile2",
        url=u"decor/snow/tile2.svg",
        height=100,
        width=100,
        theme=get_theme("grass"),
        pk=30,
    ),
    (u"tile2", u"farm"): Decor(
        z_index=0,
        name=u"tile2",
        url=u"decor/snow/tile2.svg",
        height=100,
        width=100,
        theme=get_theme("farm"),
        pk=31,
    ),
    (u"tile2", u"city"): Decor(
        z_index=0,
        name=u"tile2",
        url=u"decor/snow/tile2.svg",
        height=100,
        width=100,
        theme=get_theme("city"),
        pk=32,
    ),
}


def get_decor_element(name, theme):
    """ Helper method to get a decor element corresponding to the theme or a default one."""
    try:
        return DECOR_DATA[(name, theme.name)]
    except KeyError:
        for theme_object in get_all_themes():
            try:
                return DECOR_DATA[(name, theme_object.name)]
            except KeyError:
                pass
    raise KeyError


def get_all_decor():
    return DECOR_DATA.values()


def get_decor_element_by_pk(pk):
    for decor in DECOR_DATA.values():
        if decor.pk == int(pk):
            return decor
    raise KeyError


def get_decors_url(pk, request):
    return reverse("decor-detail", args={pk}, request=request)



# -*- coding: utf-8 -*-
# Code for Life
#
# Copyright (C) 2019, Ocado Innovation Limited
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ADDITIONAL TERMS – Section 7 GNU General Public Licence
#
# This licence does not grant any right, title or interest in any “Ocado” logos,
# trade names or the trademark “Ocado” or any other trademarks or domain names
# owned by Ocado Innovation Limited or the Ocado group of companies or any other
# distinctive brand features of “Ocado” as may be secured from time to time. You
# must not distribute any modification of this program using the trademark
# “Ocado” or claim any affiliation or association with Ocado or its employees.
#
# You are not authorised to use the name Ocado (or any of its trade names) or
# the names of any author or contributor in advertising or for publicity purposes
# pertaining to the distribution of this program, without the prior written
# authorisation of Ocado.
#
# Any propagation, distribution or conveyance of this program must include this
# copyright notice and these terms. You must not misrepresent the origins of this
# program; modified versions of the program must be marked as such and not
# identified as the original program.
from django.http import HttpResponse
from game.models import Level, Episode, LevelBlock, Block, LevelDecor
from game.serializers import (
    LevelListSerializer,
    EpisodeListSerializer,
    LevelDetailSerializer,
    EpisodeDetailSerializer,
    LevelBlockSerializer,
    BlockSerializer,
    LevelMapDetailSerializer,
    LevelDecorSerializer,
    LevelModeSerializer,
    LevelMapListSerializer,
)
from rest_framework.decorators import api_view
from rest_framework.response import Response
from rest_framework.reverse import reverse
from rest_framework import viewsets

from game.decor import get_all_decor, get_decor_element_by_pk, get_decors_url
from game.theme import get_all_themes, get_theme_by_pk, get_themes_url
from game.character import get_all_character, get_character_by_pk, get_characters_url


@api_view(("GET",))
def api_root(request, format=None):
    return Response(
        {
            "blocks": reverse("block-list", request=request, format=format),
            "characters": reverse("character-list", request=request, format=format),
            "decors": reverse("decor-list", request=request, format=format),
            "episodes": reverse("episode-list", request=request, format=format),
            "levels": reverse("level-list", request=request, format=format),
            "maps": reverse("map-list", request=request, format=format),
            "themes": reverse("theme-list", request=request, format=format),
        }
    )


@api_view(("GET",))
def decor_list(request, format=None):
    decors = get_all_decor()
    data = [{get_decors_url(i.pk, request)} for i in decors]
    return Response(data)


@api_view(("GET",))
def decor_detail(request, pk, format=None):
    try:
        decor = get_decor_element_by_pk(pk=pk)
    except KeyError:
        return HttpResponse(status=404)
    data = decor.__dict__.copy()
    data["theme"] = get_themes_url(data["theme"].pk, request)
    return Response(data)


@api_view(("GET",))
def level_list(request, format=None):
    levels = Level.objects.sorted_levels()
    serializer = LevelListSerializer(levels, many=True, context={"request": request})
    return Response(serializer.data)


# pk is the episode id
@api_view(("GET",))
def level_for_episode(request, pk, format=None):
    levels = Level.objects.filter(episode__id=pk)
    serializer = LevelListSerializer(levels, many=True, context={"request": request})
    return Response(serializer.data)


@api_view(("GET",))
def level_detail(request, pk, format=None):
    try:
        level = Level.objects.get(pk=pk)
    except Level.DoesNotExist:
        return HttpResponse(status=404)

    serializer = LevelDetailSerializer(level, context={"request": request})
    return Response(serializer.data)


@api_view(("GET",))
def map_list(request, format=None):
    levels = Level.objects.sorted_levels()

    serializer = LevelMapListSerializer(levels, many=True, context={"request": request})
    return Response(serializer.data)


@api_view(("GET",))
def map_for_level(request, pk, format=None):
    try:
        level = Level.objects.get(pk=pk)
    except Level.DoesNotExist:
        return HttpResponse(status=404)

    serializer = LevelMapDetailSerializer(level, context={"request": request})
    return Response(serializer.data)


@api_view(("GET",))
def mode_for_level(request, pk, format=None):
    try:
        level = Level.objects.get(pk=pk)
    except Level.DoesNotExist:
        return HttpResponse(status=404)

    serializer = LevelModeSerializer(level, context={"request": request})
    return Response(serializer.data)


@api_view(("GET",))
def episode_list(request, format=None):
    episodes = Episode.objects.all()
    serializer = EpisodeListSerializer(
        episodes, many=True, context={"request": request}
    )
    return Response(serializer.data)


@api_view(("GET",))
def episode_detail(request, pk, format=None):
    try:
        episode = Episode.objects.get(pk=pk)
    except Episode.DoesNotExist:
        return HttpResponse(status=404)

    serializer = EpisodeDetailSerializer(episode, context={"request": request})
    return Response(serializer.data)


@api_view(("GET",))
def levelblock_list(request, level, format=None):
    blocks = LevelBlock.objects.filter(level__id=level)

    serializer = LevelBlockSerializer(blocks, many=True, context={"request": request})
    return Response(serializer.data)


@api_view(("GET",))
def levelblock_for_level(request, pk, format=None):
    levelblocks = LevelBlock.objects.filter(level__id=pk)
    serializer = LevelBlockSerializer(
        levelblocks, many=True, context={"request": request}
    )
    return Response(serializer.data)


@api_view(("GET",))
def levelblock_detail(request, pk, format=None):
    try:
        levelblock = LevelBlock.objects.get(pk=pk)
    except LevelBlock.DoesNotExist:
        return HttpResponse(status=404)

    serializer = LevelBlockSerializer(levelblock, context={"request": request})
    return Response(serializer.data)


@api_view(("GET",))
def leveldecor_list(request, level, format=None):
    leveldecors = LevelDecor.objects.filter(level__id=level)

    serializer = LevelDecorSerializer(
        leveldecors, many=True, context={"request": request}
    )
    return Response(serializer.data)


@api_view(("GET",))
def leveldecor_for_level(request, pk, format=None):
    leveldecors = LevelDecor.objects.filter(level__id=pk)
    serializer = LevelDecorSerializer(
        leveldecors, many=True, context={"request": request}
    )
    return Response(serializer.data)


@api_view(("GET",))
def leveldecor_detail(request, pk, format=None):
    try:
        leveldecor = LevelDecor.objects.get(pk=pk)
    except LevelDecor.DoesNotExist:
        return HttpResponse(status=404)

    serializer = LevelDecorSerializer(leveldecor, context={"request": request})
    return Response(serializer.data)


@api_view(("GET",))
def block_list(request, format=None):
    block = Block.objects.all()
    serializer = BlockSerializer(block, many=True, context={"request": request})
    return Response(serializer.data)


@api_view(("GET",))
def block_detail(request, pk, format=None):
    try:
        block = Block.objects.get(pk=pk)
    except Block.DoesNotExist:
        return HttpResponse(status=404)

    serializer = BlockSerializer(block, context={"request": request})
    return Response(serializer.data)


@api_view(("GET",))
def theme_list(request, format=None):
    themes = get_all_themes()
    data = [{get_themes_url(i.pk, request)} for i in themes]
    return Response(data)


@api_view(("GET",))
def theme_detail(request, pk, format=None):
    try:
        theme = get_theme_by_pk(pk)
    except KeyError:
        return HttpResponse(status=404)
    return Response(theme.__dict__)


@api_view(("GET",))
def character_list(request, format=None):
    characters = get_all_character()
    data = [{get_characters_url(i.pk, request)} for i in characters]
    return Response(data)


@api_view(("GET",))
def character_detail(request, pk, format=None):
    try:
        character = get_character_by_pk(pk)
    except KeyError:
        return HttpResponse(status=404)
    return Response(character.__dict__)


# Maybe used later for when we use a viewset which requires multiple serializer
class MultiSerializerViewSet(viewsets.ModelViewSet):
    serializers = {"default": None}

    def get_serializer_class(self):
        return self.serializers.get(self.action, self.serializers["default"])



# -*- coding: utf-8 -*-
import re
import sys

from setuptools import find_packages, setup

with open("game/__init__.py", "r") as fd:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE
    ).group(1)

try:
    from semantic_release import setup_hook

    setup_hook(sys.argv)
except ImportError:
    pass

setup(
    name="rapid-router",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "django>=1.10.8, <= 1.11.24",
        "django-foundation-icons==3.1",
        "django-bourbon==3.1.8",
        "django-autoconfig >= 0.3.6, < 1.0.0",
        "django-jquery==1.9.1",
        "django-js-reverse==0.9.1",
        "django-foundation-statics==5.4.7",
        "django-pipeline==1.6.14",
        "django-appconf==1.0.1",
        "django-casper==0.0.3",
        "djangorestframework>=3.8.2, <3.9.0",
        "six==1.11.0",
        "docutils==0.12",
        "pyhamcrest==1.8.3",
        "libsass",
    ],
    version=version,
    zip_safe=False,
)



# -*- coding: utf-8 -*-
# Code for Life
#
# Copyright (C) 2019, Ocado Innovation Limited
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ADDITIONAL TERMS – Section 7 GNU General Public Licence
#
# This licence does not grant any right, title or interest in any “Ocado” logos,
# trade names or the trademark “Ocado” or any other trademarks or domain names
# owned by Ocado Innovation Limited or the Ocado group of companies or any other
# distinctive brand features of “Ocado” as may be secured from time to time. You
# must not distribute any modification of this program using the trademark
# “Ocado” or claim any affiliation or association with Ocado or its employees.
#
# You are not authorised to use the name Ocado (or any of its trade names) or
# the names of any author or contributor in advertising or for publicity purposes
# pertaining to the distribution of this program, without the prior written
# authorisation of Ocado.
#
# Any propagation, distribution or conveyance of this program must include this
# copyright notice and these terms. You must not misrepresent the origins of this
# program; modified versions of the program must be marked as such and not
# identified as the original program.
from datetime import timedelta, datetime
from django.utils.timezone import utc

from django.core.urlresolvers import reverse
from django.test import TestCase, Client
from hamcrest import *

from game.models import Level, Attempt
from game.views.scoreboard import StudentRow, scoreboard_data
from game.views.scoreboard_csv import (
    scoreboard_csv_multiple_levels,
    scoreboard_csv_single_level,
)
from portal.models import Class
from portal.tests.utils.classes import create_class_directly
from portal.tests.utils.organisation import create_organisation_directly
from portal.tests.utils.student import create_school_student_directly
from portal.tests.utils.teacher import signup_teacher_directly

Headers = [
    "Class",
    "Name",
    "Total Score",
    "Total Time",
    "Started Levels %",
    "Attempted levels %",
    "Finished levels %",
]


class ScoreboardTestCase(TestCase):
    def test_teacher_multiple_students_multiple_levels(self):
        level_ids = ids_of_levels_named(["1", "2"])
        level1 = Level.objects.get(name="1")
        level2 = Level.objects.get(name="2")

        clas, student, student2 = set_up_data()

        create_attempt(student, level1, 10.5)
        create_attempt(student2, level1, 2.3)
        create_attempt(student2, level2, 16.7)

        student_data, headers = scoreboard_data(Teacher(), level_ids, [clas.id])

        assert_that(
            headers,
            equal_to(
                [
                    "Class",
                    "Name",
                    "Total Score",
                    "Total Time",
                    "Progress",
                    u"Level 1",
                    u"Level 2",
                ]
            ),
        )
        assert_that(student_data, has_length(2))

        assert_student_row(
            student_row=student_data[0],
            class_name=clas.name,
            student_name=student.user.user.first_name,
            total_score=10.5,
            total_time=timedelta(0),
            progress=(0.0, 0.0, 50.0),
            scores=[10.5, ""],
        )

        assert_student_row(
            student_row=student_data[1],
            class_name=clas.name,
            student_name=student2.user.user.first_name,
            total_score=19.0,
            total_time=timedelta(0),
            progress=(0.0, 50.0, 50.0),
            scores=[2.3, 16.7],
        )

    def test_teacher_multiple_students_single_level(self):
        level_ids = ids_of_levels_named(["1"])
        level1 = Level.objects.get(name="1")

        clas, student, student2 = set_up_data()

        create_attempt(student, level1, 10.5)
        create_attempt(student2, level1, 2.3)

        student_data, headers = scoreboard_data(Teacher(), level_ids, [clas.id])

        assert_that(
            headers,
            equal_to(
                ["Class", "Name", "Score", "Total Time", "Start Time", "Finish Time"]
            ),
        )
        assert_that(student_data, has_length(2))

        assert_student_row_single_level(
            student_row=student_data[0],
            class_name=clas.name,
            student_name=student.user.user.first_name,
            total_score=10.5,
            total_time=timedelta(0),
        )

        assert_student_row_single_level(
            student_row=student_data[1],
            class_name=clas.name,
            student_name=student2.user.user.first_name,
            total_score=2.3,
            total_time=timedelta(0),
        )

    def test_student_multiple_students_multiple_levels(self):
        level_ids = ids_of_levels_named(["1", "2"])
        level1 = Level.objects.get(name="1")
        level2 = Level.objects.get(name="2")

        clas, student, student2 = set_up_data(True)

        create_attempt(student, level1, 10.5)
        create_attempt(student2, level1, 2.3)
        create_attempt(student2, level2, 16.7)

        student_data, headers = scoreboard_data(Student(student), level_ids, [clas.id])

        assert_that(
            headers,
            equal_to(
                [
                    "Class",
                    "Name",
                    "Total Score",
                    "Total Time",
                    "Progress",
                    u"Level 1",
                    u"Level 2",
                ]
            ),
        )
        assert_that(student_data, has_length(2))

        assert_student_row(
            student_row=student_data[0],
            class_name=clas.name,
            student_name=student.user.user.first_name,
            total_score=10.5,
            total_time=timedelta(0),
            progress=(0.0, 0.0, 50.0),
            scores=[10.5, ""],
        )

        assert_student_row(
            student_row=student_data[1],
            class_name=clas.name,
            student_name=student2.user.user.first_name,
            total_score=19.0,
            total_time=timedelta(0),
            progress=(0.0, 50.0, 50.0),
            scores=[2.3, 16.7],
        )

    def test_student_multiple_students_single_level(self):
        level_ids = ids_of_levels_named(["2"])
        level2 = Level.objects.get(name="2")

        clas, student, student2 = set_up_data(True)

        create_attempt(student, level2, 10.5)
        create_attempt(student2, level2, 16.7)

        student_data, headers = scoreboard_data(Student(student), level_ids, [clas.id])

        assert_that(
            headers,
            equal_to(
                ["Class", "Name", "Score", "Total Time", "Start Time", "Finish Time"]
            ),
        )
        assert_that(student_data, has_length(2))

        assert_student_row_single_level(
            student_row=student_data[0],
            class_name=clas.name,
            student_name=student.user.user.first_name,
            total_score=10.5,
            total_time=timedelta(0),
        )

        assert_student_row_single_level(
            student_row=student_data[1],
            class_name=clas.name,
            student_name=student2.user.user.first_name,
            total_score=16.7,
            total_time=timedelta(0),
        )

    def test_student_multiple_students_multiple_levels_cannot_see_classmates(self):
        level_ids = ids_of_levels_named(["1", "2"])
        level1 = Level.objects.get(name="1")
        level2 = Level.objects.get(name="2")

        clas, student, student2 = set_up_data()
        create_attempt(student, level1, 10.5)
        create_attempt(student2, level1, 2.3)
        create_attempt(student2, level2, 16.7)

        student_data, headers = scoreboard_data(Student(student), level_ids, [clas.id])

        assert_that(
            headers,
            equal_to(
                [
                    "Class",
                    "Name",
                    "Total Score",
                    "Total Time",
                    "Progress",
                    u"Level 1",
                    u"Level 2",
                ]
            ),
        )
        assert_that(student_data, has_length(1))

        assert_student_row(
            student_row=student_data[0],
            class_name=clas.name,
            student_name=student.user.user.first_name,
            total_score=10.5,
            total_time=timedelta(0),
            progress=(0.0, 0.0, 50.0),
            scores=[10.5, ""],
        )

    def test_scoreboard_loads(self):
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        klass, name, access_code = create_class_directly(email)
        create_school_student_directly(access_code)
        url = reverse("scoreboard")
        c = Client()
        c.login(username=email, password=password)
        response = c.get(url)
        self.assertEqual(200, response.status_code)


class ScoreboardCsvTestCase(TestCase):
    def test_multiple_levels(self):
        levels = Level.objects.sorted_levels()
        student_rows = [(self.student_row("secrète")), (self.student_row())]

        response = scoreboard_csv_multiple_levels(student_rows, levels)

        actual_header, actual_rows = self.actual_data(response.content)

        expected_header = self.expected_header(levels)
        expected_rows = self.expected_rows_multiple_levels(student_rows)

        assert_that(actual_header, equal_to(expected_header))

        assert_that(actual_rows, equal_to(expected_rows))

    def test_single_level(self):
        student_rows = [(self.student_row("secrète")), (self.student_row())]
        response = scoreboard_csv_single_level(student_rows)

        actual_header, actual_rows = self.actual_data(response.content)

        expected_header = "Class,Name,Score,Total Time,Start Time,Finish Time"
        expected_rows = self.expected_rows_single_level(student_rows)

        assert_that(actual_header, equal_to(expected_header))

        assert_that(actual_rows, equal_to(expected_rows))

    def expected_rows_single_level(self, student_rows):
        return map(self.expected_row_single_level, student_rows) + [""]

    def expected_rows_multiple_levels(self, student_rows):
        return map(self.expected_row_multiple_levels, student_rows) + [""]

    def student_row(self, class_name=None):
        email, password = signup_teacher_directly()
        _, class_name, access_code = create_class_directly(email, class_name)
        _, _, student = create_school_student_directly(access_code)

        total_time = timedelta(0, 30)
        scores = [x for x in range(20)]
        total_score = sum(scores)
        progress = (0, 0, 0)

        all_scores = scores + [""] * 89

        row = StudentRow(
            student=student,
            total_time=total_time,
            total_score=total_score,
            start_time=datetime.fromtimestamp(1435305072, tz=utc),
            finish_time=datetime.fromtimestamp(1438305072, tz=utc),
            progress=progress,
            scores=all_scores,
            class_field=Class(name="MyClass"),
        )
        return row

    def expected_row_multiple_levels(self, student_row):
        beginning = (
            "%s,%s,190,0:00:30,0,0,0,0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,"
            % (student_row.class_field.name.encode("utf-8"), student_row.name)
        )
        padding = ",".join([""] * 89)
        return beginning + padding

    def expected_row_single_level(self, student_row):
        return (
            "%s,%s,190,0:00:30,2015-06-26 07:51:12+00:00,2015-07-31 01:11:12+00:00"
            % (student_row.class_field.name.encode("utf-8"), student_row.name)
        )

    def expected_header(self, levels):
        level_strings = map(str, levels)
        all_header_strings = Headers + level_strings
        joined = ",".join(all_header_strings)
        return joined

    def actual_data(self, content):
        split = content.split("\r\n")
        header = split[0]
        rows = split[1:]
        return header, rows


class Student:
    def __init__(self, student):
        self.student = student

    def is_student(self):
        return True

    def is_teacher(self):
        return False

    def is_independent_student(self):
        return False


class Teacher:
    def is_student(self):
        return False

    def is_teacher(self):
        return True

    def is_independent_student(self):
        return False


def assert_student_row(
    student_row, class_name, student_name, total_score, total_time, progress, scores
):
    assert_that(student_row.class_field.name, equal_to(class_name))
    assert_that(student_row.name, equal_to(student_name))
    assert_that(student_row.total_score, equal_to(total_score))
    assert_that(student_row.total_time, equal_to(total_time))
    assert_that(student_row.progress, equal_to(progress))
    assert_that(student_row.scores, equal_to(scores))


def assert_student_row_single_level(
    student_row, class_name, student_name, total_score, total_time
):
    assert_that(student_row.class_field.name, equal_to(class_name))
    assert_that(student_row.name, equal_to(student_name))
    assert_that(student_row.total_score, equal_to(total_score))
    assert_that(student_row.total_time, equal_to(total_time))


def create_attempt(student, level, score):
    attempt = Attempt.objects.create(
        finish_time=datetime.fromtimestamp(1435305072),
        level=level,
        student=student,
        score=score,
        is_best_attempt=True,
    )
    attempt.start_time = datetime.fromtimestamp(1435305072)
    attempt.save()


def ids_of_levels_named(names):
    levels = Level.objects.filter(name__in=names)
    assert_that(len(levels), equal_to(len(names)))
    level_ids = map(lambda x: x.id, levels)
    return level_ids


def set_up_data(classmates_data_viewable=False):
    email, password = signup_teacher_directly()
    clas, class_name, access_code = create_class_directly(email)
    if classmates_data_viewable:
        clas.classmates_data_viewable = True
        clas.save()
    _, _, student = create_school_student_directly(access_code)
    _, _, student2 = create_school_student_directly(access_code)
    create_random_school_data()
    return clas, student, student2


def create_random_school_data():
    email, password = signup_teacher_directly()
    clas, class_name, access_code = create_class_directly(email)
    create_school_student_directly(access_code)
    create_school_student_directly(access_code)
